import json
import logging

import requests

from mdps_ds_lib.lib.utils.json_validator import JsonValidator

LOGGER = logging.getLogger()

COGNITO_RESULT_SCHEMA = {
    'type': 'object',
    'required': ['AuthenticationResult'],
    'properties': {
        'AuthenticationResult': {
            'type': 'object',
            'required': ['AccessToken', 'ExpiresIn', 'IdToken'],
            'properties': {
                'AccessToken': {'type': 'string'},
                'ExpiresIn': {'type': 'number'},
                'IdToken': {'type': 'string'},
                'RefreshToken': {'type': 'string'},
                'TokenType': {'type': 'string'},
            }
        }
    }
}


class CognitoLogin:
    def __init__(self):
        self.__verify_ssl = True
        self.__client_id = None
        self.__cognito_url = None
        self.__token = ''

    @property
    def verify_ssl(self):
        return self.__verify_ssl

    @verify_ssl.setter
    def verify_ssl(self, val):
        """
        :param val:
        :return: None
        """
        self.__verify_ssl = val
        return

    @property
    def client_id(self):
        return self.__client_id

    @client_id.setter
    def client_id(self, val):
        """
        :param val:
        :return: None
        """
        self.__client_id = val
        return

    @property
    def cognito_url(self):
        return self.__cognito_url

    @cognito_url.setter
    def cognito_url(self, val):
        """
        :param val:
        :return: None
        """
        self.__cognito_url = val
        return

    @property
    def token(self):
        return self.__token

    @token.setter
    def token(self, val):
        """
        :param val:
        :return: None
        """
        self.__token = val
        return

    def with_client_id(self, client_id: str):
        self.client_id = client_id
        return self

    def with_cognito_url(self, cognito_url: str):
        self.cognito_url = cognito_url
        return self

    def with_verify_ssl(self, verify_ssl: bool):
        self.verify_ssl = verify_ssl
        return self

    def start(self, username: str, dwssap: str):
        """
curl -X POST --data @cognito.jpl.aws.json
-H 'X-Amz-Target: AWSCognitoIdentityProviderService.InitiateAuth'
-H 'Content-Type: application/x-amz-json-1.1' "https://cognito-idp.us-west-2.amazonaws.com/"|jq


        :param username:
        :param dwssap:
        :return:
        """
        if self.__client_id is None or self.__cognito_url is None:
            raise ValueError(f'missing client_id or cognito_url. set them first')
        header = {
            'X-Amz-Target': 'AWSCognitoIdentityProviderService.InitiateAuth',
            'Content-Type': 'application/x-amz-json-1.1',
        }
        login_body = {
            'AuthParameters': {
                'USERNAME': username,
                'PASSWORD': dwssap,
            },
            'AuthFlow': 'USER_PASSWORD_AUTH',
            'ClientId': self.__client_id
        }
        response = requests.post(url=self.__cognito_url, headers=header, verify=self.__verify_ssl,
                                 data=json.dumps(login_body))
        if response.status_code > 400:
            raise RuntimeError(
                f'Cognito ends in error. status_code: {response.status_code}. url: {self.__cognito_url}. details: {response.text}')
        login_result = json.loads(response.content.decode('utf-8'))
        result = JsonValidator(COGNITO_RESULT_SCHEMA).validate(login_result)
        if result is not None:
            raise ValueError(f'cognito login response has validation errors: {result}')
        self.__token = login_result['AuthenticationResult']['AccessToken']
        return self



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os

import boto3


LOGGER = logging.getLogger(__name__)


class AwsCred:
    __DEFAULT_REGION = 'us-west-2'

    def __init__(self):
        self.__region = os.environ.get('AWS_REGION', AwsCred.__DEFAULT_REGION)
        LOGGER.debug(f'using region: {self.__region}')
        self.__boto3_session = {'region_name': self.__region}
        aws_access_key_id = os.environ.get('AWS_ACCESS_KEY_ID', '')
        aws_secret_access_key = os.environ.get('AWS_SECRET_ACCESS_KEY', '')
        aws_session_token = os.environ.get('AWS_SESSION_TOKEN', '')
        if aws_access_key_id != '':
            LOGGER.debug('using aws_access_key_id as it is not empty')
            if aws_secret_access_key == '':
                raise ValueError(f'missing aws_secret_access_key for aws_access_key_id ends with {aws_access_key_id[-3:]}')
            self.__boto3_session['aws_access_key_id'] = aws_access_key_id
            self.__boto3_session['aws_secret_access_key'] = aws_secret_access_key
            if aws_session_token != '':
                LOGGER.debug('adding aws_session_token as it is not empty and aws_access_key_id exists')
                self.__boto3_session['aws_session_token'] = aws_session_token
        else:
            LOGGER.debug('using default session as there is  no aws_access_key_id')

    @property
    def region(self):
        return self.__region

    @region.setter
    def region(self, val):
        """
        :param val:
        :return: None
        """
        self.__region = val
        return

    @property
    def boto3_session(self):
        return self.__boto3_session

    @boto3_session.setter
    def boto3_session(self, val):
        """
        :param val:
        :return: None
        """
        self.__boto3_session = val
        return

    def get_session(self):
        return boto3.Session(**self.boto3_session)

    def get_resource(self, service_name: str):
        return boto3.Session(**self.boto3_session).resource(service_name)

    def get_client(self, service_name: str):
        return boto3.Session(**self.boto3_session).client(service_name)



import json
import logging
import os
from typing import Union

from mdps_ds_lib.lib.utils.file_utils import FileUtils

LOGGER = logging.getLogger(__name__)


class StageInOutUtils:
    OUTPUT_FILE = 'OUTPUT_FILE'
    CATALOG_DEFAULT_CHUNK_SIZE = 5

    @staticmethod
    def write_output_to_file(output_json: Union[dict, str, list]):
        if StageInOutUtils.OUTPUT_FILE not in os.environ:
            LOGGER.debug(f'Not writing output to file due to missing {StageInOutUtils.OUTPUT_FILE} in ENV')
            return
        output_filepath = os.environ.get(StageInOutUtils.OUTPUT_FILE)
        FileUtils.mk_dir_p(os.path.dirname(output_filepath))
        output_str = json.dumps(output_json) if not isinstance(output_json, str) else output_json
        with open(output_filepath, 'w') as ff:
            ff.write(output_str)
        return

    @staticmethod
    def chunk_list(input_list, chunked_size):
        """Yield successive n-sized chunks from l."""
        for i in range(0, len(input_list), chunked_size):
            yield input_list[i:i + chunked_size]
        return



from mdps_ds_lib.lib.cumulus_stac.item_transformer import ItemTransformer


class StacUtils:
    @staticmethod
    def reduce_stac_list_to_data_links(input_stac_granules_list: [dict]):
        """
        :param input_stac_granules_list: list of pystac dictionaries
        :return: [dict] - list of assets which has data key
        """
        pystac_items = [ItemTransformer().from_stac(k) for k in input_stac_granules_list]
        assets = [k.get_assets() for k in pystac_items]
        data_only_assets = []
        for each_granule_asset in assets:
            data_only_asset = {k: v.to_dict() for k, v in each_granule_asset.items() if v.roles is not None and 'data' in v.roles}
            data_only_assets.append(data_only_asset)
        return data_only_assets



import shutil

import requests

from mdps_ds_lib.stage_in_out.download_granules_abstract import DownloadGranulesAbstract
import logging
import os

LOGGER = logging.getLogger(__name__)


class DownloadGranulesHttp(DownloadGranulesAbstract):

    def __init__(self) -> None:
        super().__init__()

    def _set_props_from_env(self):
        missing_keys = [k for k in [self.STAC_JSON, self.DOWNLOAD_DIR_KEY] if k not in os.environ]
        if len(missing_keys) > 0:
            raise ValueError(f'missing environment keys: {missing_keys}')
        self._retrieve_stac_json()
        self._setup_download_dir()
        return self

    def _download_one_item(self, downloading_url):
        downloading_response = requests.get(downloading_url, stream=True)
        downloading_response.raise_for_status()
        downloading_response.raw.decode_content = True
        local_file_path = os.path.join(self._download_dir, os.path.basename(downloading_url))
        with open(local_file_path, 'wb') as f:
            shutil.copyfileobj(downloading_response.raw, f)
        return local_file_path



import json
import logging
from urllib.parse import quote_plus, urlparse, unquote_plus

import pystac
from pystac.utils import datetime_to_str

from pystac import Link, Collection, Extent, SpatialExtent, TemporalExtent, Summaries

from mdps_ds_lib.lib.constants import Constants
from mdps_ds_lib.lib.cumulus_stac.stac_transformer_abstract import StacTransformerAbstract
from mdps_ds_lib.lib.utils.time_utils import TimeUtils

STAC_COLLECTION_SCHEMA = '''{
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$id": "https://schemas.stacspec.org/v1.0.0/collection-spec/json-schema/collection.json#",
    "title": "STAC Collection Specification",
    "description": "This object represents Collections in a SpatioTemporal Asset Catalog.",
    "allOf": [
        {
            "$ref": "#/definitions/collection"
        }
    ],
    "definitions": {
        "collection": {
            "title": "STAC Collection",
            "description": "These are the fields specific to a STAC Collection. All other fields are inherited from STAC Catalog.",
            "type": "object",
            "required": [
                "stac_version",
                "type",
                "id",
                "description",
                "license",
                "extent",
                "links"
            ],
            "properties": {
                "stac_version": {
                    "title": "STAC version",
                    "type": "string",
                    "const": "1.0.0"
                },
                "stac_extensions": {
                    "title": "STAC extensions",
                    "type": "array",
                    "uniqueItems": true,
                    "items": {
                        "title": "Reference to a JSON Schema",
                        "type": "string",
                        "format": "iri"
                    }
                },
                "type": {
                    "title": "Type of STAC entity",
                    "const": "Collection"
                },
                "id": {
                    "title": "Identifier",
                    "type": "string",
                    "minLength": 1
                },
                "title": {
                    "title": "Title",
                    "type": "string"
                },
                "description": {
                    "title": "Description",
                    "type": "string",
                    "minLength": 1
                },
                "keywords": {
                    "title": "Keywords",
                    "type": "array",
                    "items": {
                        "type": "string"
                    }
                },
                "license": {
                    "title": "Collection License Name",
                    "type": "string",
                    "pattern": "^[\\\\w\\\\-\\\\.\\\\+]+$"
                },
                "providers": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "required": [
                            "name"
                        ],
                        "properties": {
                            "name": {
                                "title": "Organization name",
                                "type": "string"
                            },
                            "description": {
                                "title": "Organization description",
                                "type": "string"
                            },
                            "roles": {
                                "title": "Organization roles",
                                "type": "array",
                                "items": {
                                    "type": "string",
                                    "enum": [
                                        "producer",
                                        "licensor",
                                        "processor",
                                        "host"
                                    ]
                                }
                            },
                            "url": {
                                "title": "Organization homepage",
                                "type": "string",
                                "format": "iri"
                            }
                        }
                    }
                },
                "extent": {
                    "title": "Extents",
                    "type": "object",
                    "required": [
                        "spatial",
                        "temporal"
                    ],
                    "properties": {
                        "spatial": {
                            "title": "Spatial extent object",
                            "type": "object",
                            "required": [
                                "bbox"
                            ],
                            "properties": {
                                "bbox": {
                                    "title": "Spatial extents",
                                    "type": "array",
                                    "minItems": 1,
                                    "items": {
                                        "title": "Spatial extent",
                                        "type": "array",
                                        "oneOf": [
                                            {
                                                "minItems": 4,
                                                "maxItems": 4
                                            },
                                            {
                                                "minItems": 6,
                                                "maxItems": 6
                                            }
                                        ],
                                        "items": {
                                            "type": "number"
                                        }
                                    }
                                }
                            }
                        },
                        "temporal": {
                            "title": "Temporal extent object",
                            "type": "object",
                            "required": [
                                "interval"
                            ],
                            "properties": {
                                "interval": {
                                    "title": "Temporal extents",
                                    "type": "array",
                                    "minItems": 1,
                                    "items": {
                                        "title": "Temporal extent",
                                        "type": "array",
                                        "minItems": 2,
                                        "maxItems": 2,
                                        "items": {
                                            "type": [
                                                "string",
                                                "null"
                                            ],
                                            "format": "date-time",
                                            "pattern": "(\\\\+00:00|Z)$"
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "assets": {
                    "$ref": "../../item-spec/json-schema/item.json#/definitions/assets"
                },
                "links": {
                    "title": "Links",
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/link"
                    }
                },
                "summaries": {
                    "$ref": "#/definitions/summaries"
                }
            }
        },
        "link": {
            "type": "object",
            "required": [
                "rel",
                "href"
            ],
            "properties": {
                "href": {
                    "title": "Link reference",
                    "type": "string",
                    "format": "iri-reference",
                    "minLength": 1
                },
                "rel": {
                    "title": "Link relation type",
                    "type": "string",
                    "minLength": 1
                },
                "type": {
                    "title": "Link type",
                    "type": "string"
                },
                "title": {
                    "title": "Link title",
                    "type": "string"
                }
            }
        },
        "summaries": {
            "type": "object",
            "additionalProperties": {
                "anyOf": [
                    {
                        "title": "JSON Schema",
                        "type": "object",
                        "minProperties": 1,
                        "allOf": [
                            {
                                "$ref": "http://json-schema.org/draft-07/schema"
                            }
                        ]
                    },
                    {
                        "title": "Range",
                        "type": "object",
                        "required": [
                            "minimum",
                            "maximum"
                        ],
                        "properties": {
                            "minimum": {
                                "title": "Minimum value",
                                "type": [
                                    "number",
                                    "string"
                                ]
                            },
                            "maximum": {
                                "title": "Maximum value",
                                "type": [
                                    "number",
                                    "string"
                                ]
                            }
                        }
                    },
                    {
                        "title": "Set of values",
                        "type": "array",
                        "minItems": 1
                    }
                ]
            }
        }
    }
}'''
# The following lines are a part of summaries > additionalProperties > oneOf > [last item].
# But including that will not make it a valid schema.
# excluding it for now
# "items": {
#     "description": "For each field only the original data type of the property can occur (except for arrays), but we can't validate that in JSON Schema yet. See the sumamry description in the STAC specification for details."
# }
LOGGER = logging.getLogger(__name__)


class CollectionTransformer(StacTransformerAbstract):
    def __init__(self, report_to_ems:bool = True, include_date_range=False, items_base_url=''):
        super().__init__()
        self.__stac_collection_schema = json.loads(STAC_COLLECTION_SCHEMA)
        self.__cumulus_collection_schema = {}
        self.__report_to_ems = report_to_ems
        self.__include_date_range = include_date_range

        self.__output_provider = None
        self.__output_cumulus_collection = None
        self.__input_dapa_collection: Collection = None
        self.__items_base_url = items_base_url

    def get_collection_id(self):
        if self.__input_dapa_collection is None:
            raise ValueError(f'pls load and parse __input_dapa_collection via from_stac')
        return self.__input_dapa_collection.id

    def get_collection_bbox(self):
        if self.__input_dapa_collection is None:
            raise ValueError(f'pls load and parse __input_dapa_collection via from_stac')
        return self.__input_dapa_collection.extent.spatial.bboxes[0]

    def get_collection_time_range(self):
        if self.__input_dapa_collection is None:
            raise ValueError(f'pls load and parse __input_dapa_collection via from_stac')
        return self.__input_dapa_collection.extent.temporal.intervals

    @property
    def output_provider(self):
        return self.__output_provider

    def generate_target_link_url(self, regex: str = None, bucket: str = None):
        href_link = ['unknown_bucket', 'unknown_regex']
        if regex is not None and regex != '':
            href_link[1] = regex
        if bucket is not None and bucket != '':
            href_link[0] = bucket
        return f"./collection.json?bucket={href_link[0]}&regex={quote_plus(href_link[1])}"

    def __convert_to_stac_link_obj(self, collection_file_obj: dict, rel_type: str = 'item'):
        """
        expected output
        {
          "title": "<sampleFileName>",
          "href": "<bucket>___<regex>"
          "type": "<type>",
          "rel": "item"
        }

        Sample input:
                {
                    "bucket": "internal",
                    "regex": "^P[0-9]{3}[0-9]{4}[A-Z]{13}T[0-9]{12}00\\.PDS$",
                    "sampleFileName": "P1570515ATMSSCIENCEAXT11344000000000.PDS",
                    "type": "data",
                    "reportToEms": True
                }
        TODO: missing reportToEms
        :param collection_file_obj:
        :return: dict
        """
        temp_link = Link(target=self.generate_target_link_url(
                collection_file_obj['regex'] if 'regex' in collection_file_obj else None,
                collection_file_obj['bucket'] if 'bucket' in collection_file_obj else None,
            ),
            rel=rel_type
            )
        if 'type' in collection_file_obj:
            temp_link.media_type = collection_file_obj['type']
        if 'sampleFileName' in collection_file_obj:
            temp_link.title = collection_file_obj['sampleFileName']
        return temp_link

    def to_stac(self, source: dict) -> dict:
        source_sample = {
            "createdAt": 1647992847582,
            "granuleId": "^P[0-9]{3}[0-9]{4}[A-Z]{13}T[0-9]{12}0$",
            "process": "modis",
            "sampleFileName": "P1570515ATMSSCIENCEAXT11344000000001.PDS",
            "name": "ATMS_SCIENCE_Group",
            "files": [
                {
                    "bucket": "internal",
                    "regex": "^P[0-9]{3}[0-9]{4}[A-Z]{13}T[0-9]{12}00\\.PDS$",
                    "sampleFileName": "P1570515ATMSSCIENCEAXT11344000000000.PDS",
                    "type": "data",
                    "reportToEms": True
                },
                {
                    "bucket": "internal",
                    "regex": "^P[0-9]{3}[0-9]{4}[A-Z]{13}T[0-9]{12}01\\.PDS$",
                    "sampleFileName": "P1570515ATMSSCIENCEAXT11344000000001.PDS",
                    "reportToEms": True,
                    "type": "metadata"
                },
                {
                    "bucket": "internal",
                    "regex": "^P[0-9]{3}[0-9]{4}[A-Z]{13}T[0-9]{12}01\\.PDS\\.xml$",
                    "sampleFileName": "P1570515ATMSSCIENCEAXT11344000000001.PDS.xml",
                    "reportToEms": True,
                    "type": "metadata"
                },
                {
                    "bucket": "internal",
                    "regex": "^P[0-9]{3}[0-9]{4}[A-Z]{13}T[0-9]{12}00.PDS.cmr.xml$",
                    "sampleFileName": "P1570515ATMSSCIENCEAXT11344000000000.PDS.cmr.xml",
                    "reportToEms": True,
                    "type": "metadata"
                }
            ],
            "granuleIdExtraction": "(P[0-9]{3}[0-9]{4}[A-Z]{13}T[0-9]{12}0).+",
            "reportToEms": True,
            "version": "001",
            "duplicateHandling": "replace",
            "updatedAt": 1647992847582,
            "url_path": "{cmrMetadata.Granule.Collection.ShortName}___{cmrMetadata.Granule.Collection.VersionId}",
            "timestamp": 1647992849273
        }
        collection_id = f"{source['name']}___{source['version']}"
        temporal_extent = []
        if 'dateFrom' in source:
            temporal_extent.append(self.get_time_obj(source['dateFrom']))
        if 'dateTo' in source:
            temporal_extent.append(self.get_time_obj(source['dateTo']))
        stac_collection = Collection(
            id=collection_id,
            # href=f"https://ideas-api-to-be-hosted/slcp/collections/{input_collection['ShortName']}::{input_collection['VersionId']}",
            description="TODO",
            extent=Extent(
                SpatialExtent([[0.0, 0.0, 0.0, 0.0]]),
                TemporalExtent([temporal_extent])
            ),
            license="proprietary",
            providers=[],
            # title=input_collection['LongName'],
            # keywords=[input_collection['SpatialKeywords']['Keyword']],
            summaries=Summaries({
                "updated": [datetime_to_str(TimeUtils().parse_from_unix(source['updatedAt'], True).get_datetime_obj())],
                "granuleId": [source['granuleId'] if 'granuleId' in source else ''],
                "granuleIdExtraction": [source['granuleIdExtraction'] if 'granuleIdExtraction' in source else ''],
                "process": [source['process'] if 'process' in source else ''],
                "totalGranules": [source['total_size'] if 'total_size' in source else -1],
            }),
            # assets={}
        )
        stac_collection.links = [self.__convert_to_stac_link_obj({
            "regex": source['url_path'] if 'url_path' in source else './collection.json',
            "sampleFileName": source['sampleFileName'],
            "type": "application/json",

        }, 'root')] + \
                                [self.__convert_to_stac_link_obj(k) for k in source['files']] + \
                                [Link(rel='parent', target=f'{self.__items_base_url}/{Constants.CATALOG}', media_type='application/json', title=f"UDS Catalog"),
                                 Link(rel='items', target=f'{self.__items_base_url}/{Constants.COLLECTIONS}/{collection_id}/items', media_type='application/json', title=f"{collection_id} Granules")]
        return stac_collection.to_dict(include_self_link=False, transform_hrefs=False)

    def get_href(self, input_href: str):
        parse_result = urlparse(input_href)
        if parse_result.query == '':
            return ''
        query_dict = [k.split('=') for k in parse_result.query.split('&')]
        query_dict = {k[0]: unquote_plus(k[1]) for k in query_dict}
        return query_dict

    def __convert_from_stac_links(self, link_obj: dict):
        output_file_object = {
            'reportToEms': self.__report_to_ems
        }
        if 'type' in link_obj:
            output_file_object['type'] = link_obj['type']
        if 'title' in link_obj:
            output_file_object['sampleFileName'] = link_obj['title']
        if 'href' in link_obj:
            href_dict = self.get_href(link_obj['href'])
            if 'bucket' in href_dict:
                output_file_object['bucket'] = href_dict['bucket']
            if 'regex' in href_dict:
                output_file_object['regex'] = href_dict['regex']
        return output_file_object

    def from_stac(self, source: dict) -> dict:
        self.__input_dapa_collection = pystac.Collection.from_dict(source)
        if not self.__input_dapa_collection.validate():
            raise ValueError(f'invalid source dapa: {self.__input_dapa_collection}')
        output_collection_cumulus = {
            # "createdAt": 1647992847582,
            "reportToEms": self.__report_to_ems,
            "duplicateHandling": "skip",
            # "updatedAt": 1647992847582,
            # "timestamp": 1647992849273
        }
        summaries = self.__input_dapa_collection.summaries.lists
        if 'granuleId' in summaries:
            output_collection_cumulus['granuleId'] = summaries['granuleId'][0]
        if 'granuleIdExtraction' in summaries:
            output_collection_cumulus['granuleIdExtraction'] = summaries['granuleIdExtraction'][0]
        if 'process' in summaries:
            output_collection_cumulus['process'] = summaries['process'][0]
        name_version = self.__input_dapa_collection.id.split('___')
        output_collection_cumulus['name'] = name_version[0]
        output_collection_cumulus['version'] = name_version[1]
        output_files = []
        for each_link_obj in self.__input_dapa_collection.links:
            each_link_obj: Link = each_link_obj
            each_file_obj = self.__convert_from_stac_links(each_link_obj.to_dict())
            if each_link_obj.rel == 'root':
                if 'regex' in each_file_obj:
                    output_collection_cumulus['url_path'] = each_file_obj['regex']
                if 'sampleFileName' in each_file_obj:
                    output_collection_cumulus['sampleFileName'] = each_file_obj['sampleFileName']
            else:
                output_files.append(each_file_obj)
        output_collection_cumulus['files'] = output_files
        if len(self.__input_dapa_collection.extent.temporal.intervals) > 0:
            date_interval = self.__input_dapa_collection.extent.temporal.intervals[0]
            if len(date_interval) == 2 and self.__include_date_range is True:
                if date_interval[0] is not None:
                    output_collection_cumulus['dateFrom'] = date_interval[0].strftime(TimeUtils.MMDD_FORMAT)
                if date_interval[1] is not None:
                    output_collection_cumulus['dateTo'] = date_interval[1].strftime(TimeUtils.MMDD_FORMAT)
        LOGGER.debug(f'self.__input_dapa_collection.providers: {self.__input_dapa_collection.providers}')
        self.__output_provider = None if self.__input_dapa_collection.providers is None or len(self.__input_dapa_collection.providers) < 1 else self.__input_dapa_collection.providers[0].name
        self.__output_cumulus_collection = output_collection_cumulus
        return output_collection_cumulus



from mdps_ds_lib.lib.utils.factory_abstract import FactoryAbstract


class ESFactory(FactoryAbstract):
    NO_AUTH = 'NO_AUTH'
    AWS = 'AWS'

    def get_instance(self, class_type, **kwargs):
        ct = class_type.upper()
        if ct == self.NO_AUTH:
            from mdps_ds_lib.lib.aws.es_middleware import ESMiddleware
            return ESMiddleware(kwargs['index'], kwargs['base_url'], port=kwargs['port'])
        if ct == self.AWS:
            from mdps_ds_lib.lib.aws.es_middleware_aws import EsMiddlewareAws
            return EsMiddlewareAws(kwargs['index'], kwargs['base_url'], port=kwargs['port'])
        raise ModuleNotFoundError(f'cannot find ES class for {ct}')



import logging
from abc import ABC, abstractmethod

from mdps_ds_lib.lib.utils.time_utils import TimeUtils

LOGGER = logging.getLogger(__name__)


class StacTransformerAbstract(ABC):
    def __init__(self) -> None:
        super().__init__()
        self._dt_formats = [
            '%y-%m-%dT%H:%M:%S.%f%z',
            '%y-%m-%dT%H:%M:%S%z',
            '%Y-%m-%dT%H:%M:%S.%f%z',
            '%Y-%m-%dT%H:%M:%S%z',
            '%Y-%m-%dT%H:%M:%S.%f',
        ]

    def get_time_obj(self, datetime_str: str):
        if datetime_str is None:
            return None
        for each_fmt in self._dt_formats:
            try:
                dt_utils = TimeUtils().parse_from_str(datetime_str, each_fmt)
                return dt_utils.get_datetime_obj()
            except ValueError as ve1:
                LOGGER.debug(f'format and value do not match: {each_fmt} v. {datetime_str}')
        raise ValueError(f'unknown format: {datetime_str}')

    @abstractmethod
    def to_stac(self, source: dict) -> dict:
        return {}

    @abstractmethod
    def from_stac(self, source: dict) -> dict:
        return {}



import json
from multiprocessing import Manager

from mdps_ds_lib.lib.utils.time_utils import TimeUtils

from mdps_ds_lib.lib.utils.file_utils import FileUtils

from pystac import ItemCollection

from mdps_ds_lib.lib.cumulus_stac.granules_catalog import GranulesCatalog
from mdps_ds_lib.lib.processing_jobs.job_executor_abstract import JobExecutorAbstract
from mdps_ds_lib.lib.processing_jobs.job_manager_abstract import JobManagerProps
from mdps_ds_lib.lib.processing_jobs.job_manager_memory import JobManagerMemory
from mdps_ds_lib.lib.processing_jobs.multithread_processor import MultiThreadProcessorProps, \
    MultiThreadProcessor
from mdps_ds_lib.stage_in_out.upload_granules_abstract import UploadGranulesAbstract
import logging
import os
from mdps_ds_lib.lib.aws.aws_s3 import AwsS3

LOGGER = logging.getLogger(__name__)


class UploadItemExecutor(JobExecutorAbstract):
    def __init__(self, result_list, error_list, collection_id, staging_bucket, retry_wait_time_sec, retry_times, delete_files: bool) -> None:
        super().__init__()
        self.__collection_id = collection_id
        self.__staging_bucket = staging_bucket
        self.__delete_files = delete_files

        self.__result_list = result_list
        self.__error_list = error_list
        self.__gc = GranulesCatalog()
        self.__s3 = AwsS3()
        self.__retry_wait_time_sec = retry_wait_time_sec
        self.__retry_times = retry_times

    def validate_job(self, job_obj):
        return True

    # def __upload_function_w_retry(self, ):
    # NOTE: 2023-10-09: we are not proceeding with "retry" logic at this moment as most the errors (when failed to upload) are not transient.
    #     upload_try_count = 1
    #     while r.status_code in [502, 504] and upload_try_count < self.__retry_times:
    #         LOGGER.error(f'502 or 504 while downloading {upload_try_count}. attempt: {upload_try_count}')
    #         time.sleep(self.__retry_wait_time_sec)
    #         upload_try_count += 1
    #     return

    def execute_job(self, each_child, lock) -> bool:
        current_granule_stac = self.__gc.get_granules_item(each_child)
        try:
            current_granules_dir = os.path.dirname(each_child)
            current_assets = self.__gc.extract_assets_href(current_granule_stac, current_granules_dir)
            if 'data' not in current_assets:  # this is still ok .coz extract_assets_href is {'data': [url1, url2], ...}
                LOGGER.warning(f'skipping {each_child}. no data in {current_assets}')
                current_granule_stac.properties['upload_error'] = f'missing "data" in assets'
                self.__error_list.put(current_granule_stac.to_dict(False, False))
                return True
            current_granule_id = str(current_granule_stac.id)
            if current_granule_id in ['', 'NA', None]:
                raise ValueError(f'invalid current_granule_id in granule {each_child}: {current_granule_id} ...')
            updating_assets = {}
            uploading_current_granule_stac = None
            for asset_type, asset_hrefs in current_assets.items():
                for each_asset_href in asset_hrefs:
                    LOGGER.debug(f'uploading {asset_type}, {each_asset_href}')
                    s3_url = self.__s3.upload(each_asset_href, self.__staging_bucket,
                                              f'{self.__collection_id}/{self.__collection_id}:{current_granule_id}',
                                              self.__delete_files)
                    if each_asset_href == each_child:
                        uploading_current_granule_stac = s3_url
                    updating_assets[os.path.basename(s3_url)] = s3_url
            self.__gc.update_assets_href(current_granule_stac, updating_assets)
            current_granule_stac.id = current_granule_id
            current_granule_stac.collection_id = self.__collection_id
            if uploading_current_granule_stac is not None:  # upload metadata file again
                self.__s3.set_s3_url(uploading_current_granule_stac)
                self.__s3.upload_bytes(json.dumps(current_granule_stac.to_dict(False, False)).encode())
            current_granule_stac.id = f'{self.__collection_id}:{current_granule_id}'
            self.__result_list.put(current_granule_stac.to_dict(False, False))
        except Exception as e:
            current_granule_stac.properties['upload_error'] = str(e)
            LOGGER.exception(f'error while processing: {each_child}')
            self.__error_list.put(current_granule_stac.to_dict(False, False))
        return True


class UploadGranulesByCompleteCatalogS3(UploadGranulesAbstract):
    CATALOG_FILE = 'CATALOG_FILE'

    def __init__(self) -> None:
        super().__init__()
        self.__gc = GranulesCatalog()
        self.__s3 = AwsS3()

    def upload(self, **kwargs) -> str:
        self._set_props_from_env()
        output_dir = os.environ.get(self.OUTPUT_DIRECTORY)
        if not FileUtils.dir_exist(output_dir):
            raise ValueError(f'OUTPUT_DIRECTORY: {output_dir} does not exist')
        missing_keys = [k for k in [self.CATALOG_FILE] if k not in os.environ]
        if len(missing_keys) > 0:
            raise ValueError(f'missing environment keys: {missing_keys}')
        catalog_file_path = os.environ.get(self.CATALOG_FILE)
        child_links = self.__gc.get_child_link_hrefs(catalog_file_path)
        local_items = Manager().Queue()
        error_list = Manager().Queue()
        job_manager_props = JobManagerProps()
        for each_child in child_links:
            job_manager_props.memory_job_dict[each_child] = each_child

        # https://www.infoworld.com/article/3542595/6-python-libraries-for-parallel-processing.html
        multithread_processor_props = MultiThreadProcessorProps(self._parallel_count)
        multithread_processor_props.job_manager = JobManagerMemory(job_manager_props)
        multithread_processor_props.job_executor = UploadItemExecutor(local_items, error_list, self._collection_id, self._staging_bucket, self._retry_wait_time_sec, self._retry_times, self._delete_files)
        multithread_processor = MultiThreadProcessor(multithread_processor_props)
        multithread_processor.start()

        LOGGER.debug(f'finished uploading all granules')
        dapa_body_granules = []
        while not local_items.empty():
            dapa_body_granules.append(local_items.get())

        errors = []
        while not error_list.empty():
            errors.append(error_list.get())
        LOGGER.debug(f'successful count: {len(dapa_body_granules)}. failed count: {len(errors)}')
        successful_item_collections = ItemCollection(items=dapa_body_granules)
        failed_item_collections = ItemCollection(items=errors)
        successful_features_file = os.path.join(output_dir, 'successful_features.json')



        failed_features_file = os.path.join(output_dir, 'failed_features.json')
        LOGGER.debug(f'writing results: {successful_features_file} && {failed_features_file}')
        FileUtils.write_json(successful_features_file, successful_item_collections.to_dict(False))
        FileUtils.write_json(failed_features_file, failed_item_collections.to_dict(False))
        s3_url = self.__s3.upload(successful_features_file, self._staging_bucket,
                                  self._result_path_prefix,
                                  s3_name=f'successful_features_{TimeUtils.get_current_time()}.json',
                                  delete_files=self._delete_files)
        LOGGER.debug(f'uploaded successful features to S3: {s3_url}')
        LOGGER.debug(f'creating response catalog')
        catalog_json = GranulesCatalog().update_catalog(catalog_file_path, [successful_features_file, failed_features_file])
        LOGGER.debug(f'catalog_json: {catalog_json}')
        return json.dumps(catalog_json)
